#include <cmath>

#include "ageinglawstateupdater.hh"
#include "../../tobool.hh"

template <class ScalarVector, class Vector>
AgeingLawStateUpdater<ScalarVector, Vector>::AgeingLawStateUpdater(
    const std::vector<ScalarVector>& _alpha_initial,
    const std::vector<Dune::BitSetVector<1>>& _nodes,
    const std::vector<double>& _L,
    const std::vector<double>& _V0)
    : alpha(_alpha_initial), nodes(_nodes), L(_L), V0(_V0) {}

template <class ScalarVector, class Vector>
void AgeingLawStateUpdater<ScalarVector, Vector>::nextTimeStep() {
  alpha_o = alpha;
}

template <class ScalarVector, class Vector>
void AgeingLawStateUpdater<ScalarVector, Vector>::setup(double _tau) {
  tau = _tau;
}

/*
  Compute [ 1-\exp(c*x) ] / x under the assumption that x is
  non-negative
*/
double liftSingularity(double c, double x) {
  if (x <= 0)
    return -c;
  else
    return -std::expm1(c * x) / x;
}

template <class ScalarVector, class Vector>
void AgeingLawStateUpdater<ScalarVector, Vector>::solve(
  const std::vector<Vector>& velocity_field) {
  for (size_t i = 0; i < alpha.size(); ++i) {
    const auto& velocity_fieldi = velocity_field[i];
    const auto& nodesi = nodes[i];
    auto& alphai = alpha[i];
    auto& alpha_oi = alpha_o[i];

    for (size_t j = 0; j < nodesi.size(); ++j) {
      if (not toBool(nodesi[j]))
        continue;

      double const V = velocity_fieldi[j].two_norm();
      double const mtoL = -tau / L[i];
      alphai[j] = std::log(std::exp(alpha_oi[j] + V * mtoL) +
                        V0[i] * liftSingularity(mtoL, V));
    }
  }
}

template <class ScalarVector, class Vector>
void AgeingLawStateUpdater<ScalarVector, Vector>::extractAlpha(
    std::vector<ScalarVector>& _alpha) {
  _alpha = alpha;
}

template <class ScalarVector, class Vector>
std::shared_ptr<StateUpdater<ScalarVector, Vector>>
AgeingLawStateUpdater<ScalarVector, Vector>::clone() const {
  return std::make_shared<AgeingLawStateUpdater<ScalarVector, Vector>>(*this);
}
